/* giFTcurs - curses interface to giFT
 * Copyright (C) 2001, 2002, 2003 Gran Weinholt <weinholt@dtek.chalmers.se>
 * Copyright (C) 2003 Christian Hggstrm <chm@c00.info>
 *
 * This file is part of giFTcurs.
 *
 * giFTcurs is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * giFTcurs is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with giFTcurs; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,  USA.
 *
 * $Id: protocol.c,v 1.36 2003/04/14 18:54:18 chnix Exp $
 */
#include "giftcurs.h"

#include <ctype.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "parse.h"
#include "protocol.h"
#include "screen.h"

/* The tree structure used by the protocol is defined here. It is
 * stored backwards, with the last items first in the linked list.
 * (and Yes, it's only to confuse you :)
 */
struct _ntree {
	char *name;
	char *value;
	ntree *next;
	ntree *subtree;
	unsigned is_integer:1;
};

static char *get_unescaped_until(char **packet, char limiter)
{
	char *str, *ptr, *txt;

	txt = ptr = str = *packet;

	while (*str != limiter) {
		if ((unsigned char) *str < ' ') {
			str++;
			continue;
		}
		if (*str == '\\')
			str++;
		if (*str == '\0')
			break;
		*ptr++ = *str++;
	}

	*packet = str + 1;

	*ptr = '\0';
	return txt;
}

static char get_string(char **packet)
{
	char *str, *txt, ch;

	txt = str = *packet;
	while (isalnum(*str) || *str == '_')
		str++;
	*packet = str;
	ch = *str;
	*str = '\0';
	return ch;
}

static ntree *parse(char **packet)
{
	ntree *key = NULL;
	char ch = **packet;

	while (1) {
		switch (ch) {
		case ' ':
		case '\t':
		case '\n':
		case '\v':
		case '\r':
			(*packet)++;
			break;
			/* Do not fret, for it is merely a wonderful gcc extension. */
		case 'A'...'Z':
		case 'a'...'z':
		case '0'...'9':
		case '_':
			interface_append(&key, *packet, NULL);
			ch = get_string(packet);
			/* ch contains the next char and **packet is overwritten with \0 */
			continue;
		case '(':
			if (!key)
				goto error;
			if (key->value)
				goto error;
			(*packet)++;
			key->value = get_unescaped_until(packet, ')');
			break;
#if 0
		case '[':
			if (!key)
				goto error;
			(*packet)++;
			get_unescaped_until(packet, ']');
			break;
#endif
		case '{':
			if (!key)
				goto error;
			(*packet)++;
			key->subtree = parse(packet);
			break;
		default:
		  error:
			DEBUG("Parse error at char %c, followed by %-30.30s", ch, *packet + 1);
		case '}':
		case ';':
			(*packet)++;
		case '\0':
			return key;
		}
		ch = **packet;
	}
}

ntree *interface_parse(char *packet)
{
	ntree *result = parse(&packet);

	if (result && result->next) {
		/* Relink so that for example SHARE[0] action(sync);
		 * becomes                    SHARE[0] { action(sync) } ;
		 * which is more tree-like */
		/* sync action->SHARE */
		/* SHARE->{ sync action } */
		ntree *last, *second_last = NULL;

		for (last = result; last->next; last = last->next)
			second_last = last;
		assert(second_last && second_last->next == last);

		last->subtree = result;
		second_last->next = NULL;

		result = last;
	}

	return result;
}

void interface_free(ntree * key)
{
	if (!key)
		return;

	interface_free(key->subtree);
	interface_free(key->next);
	free(key);
}

struct _string {
	char *str;
	int alloc;
	int len;
};

static void escape(struct _string *s, char *src)
{
	int n = s->len;
	char *dst = s->str + n;

	for (;;) {
		switch (*src) {
		case '\0':
			*dst = '\0';
			s->len = n;
			return;
		case '(':
		case ')':
		case '[':
		case ']':
		case '{':
		case '}':
		case ';':
			*dst++ = '\\';
			n++;
		default:
			*dst++ = *src++;
			n++;
		}
	}
}

static void build(ntree * key, struct _string *s)
{
	int required_space;
	char *v;

	if (!key)
		return;

	if (key->next)
		build(key->next, s);

	v = key->value;
	if (key->is_integer)
		v = itoa((unsigned int) v);

	required_space = 2 * strlen(key->name) + 32;	/* just to be sure */
	if (v)
		required_space += 2 * strlen(v) + 2;

	if (s->len + required_space > s->alloc) {
		/* gotta alloc some more. double size in each reallocation */
		s->alloc = (s->len + required_space) * 2;
		s->str = realloc(s->str, s->alloc);
		assert(s->str);
	}

	escape(s, key->name);

	if (v) {
		s->str[s->len++] = '(';
		escape(s, v);
		s->str[s->len++] = ')';
	}

	if (key->subtree) {
		s->str[s->len++] = '{';
		build(key->subtree, s);
		s->str[s->len++] = '}';
	}
	s->str[s->len++] = ' ';
}

char *interface_construct(ntree * tree)
{
	struct _string s = { NULL, 0, 0 };

	build(tree, &s);

	/* rewind the trailing space for cleanliness */
	/* also, giFT 020825 seems to _require_ no space before ; */
	s.len--;
	s.str[s.len++] = ';';
	s.str[s.len] = '\0';

	return s.str;
}

char *interface_lookup(ntree * tree, char *key)
{
	assert(key);

	/* search for the appropriate node */
	tree = tree->subtree;
	while (tree) {
		if (!strcmp(tree->name, key))
			return tree->value;
		tree = tree->next;
	}
	return NULL;
}

char *interface_name(ntree * tree)
{
	return tree->name;
}

char *interface_value(ntree * tree)
{
	return tree->value;
}

int interface_isempty(ntree * tree)
{
	return !(tree && tree->subtree);
}

void interface_append(ntree ** packet, char *key_name, char *value)
{
	ntree *key;

	assert(packet);

	key = calloc(1, sizeof(ntree));

	assert(key);

	key->name = key_name;
	key->next = *packet;
	key->value = value;

	*packet = key;
}

void interface_append_int(ntree ** packet, char *key_name, unsigned int value)
{
	interface_append(packet, key_name, (char *) value);
	(*packet)->is_integer = 1;
}

void interface_foreach(ntree * key, PForEachFunc func, void *udata)
{
	for (key = key->subtree; key; key = key->next)
		if (key->subtree)
			func(key, udata);
}

void interface_foreach_key(ntree * key, PForEachFunc func, void *udata)
{
	for (key = key->subtree; key; key = key->next)
		func(key, udata);
}
